@@ -37,4 +37,10 @@ int semanage_direct_is_managed(struct semanage_handle *sh);
 
 int semanage_direct_access_check(struct semanage_handle *sh);
 
+int semanage_direct_mls_enabled(struct semanage_handle *sh);
+
+#include <stdio.h>
+#include <unistd.h>
+ssize_t bunzip(FILE *f, char **data);
+
 #endif
