@@ -192,6 +192,11 @@ static int fcontext_matches(const semanage_fcontext_t *fcontext, void *varg)
 			goto done;
 	}
 
+	if (ustr_cmp_suffix_cstr_eq(expr, "/")) {
+		if (!ustr_del(&expr, 1))
+			goto done;
+	}
+
 	/* Append pattern to eat up trailing slashes */
 	if (!ustr_add_cstr(&expr, "/*$"))
 		goto done;
@@ -757,6 +762,10 @@ static genhomedircon_user_entry_t *get_users(genhomedircon_settings_t * s,
 		if (strcmp(name, TEMPLATE_SEUSER) == 0)
 			continue;
 
+		/* %groupname syntax */
+		if (name[0] == '%')
+			continue;
+
 		/* find the user structure given the name */
 		u = bsearch(seuname, user_list, nusers, sizeof(semanage_user_t *),
 			    (int (*)(const void *, const void *))
