@@ -60,6 +60,9 @@ typedef struct dbase_policydb dbase_t;
 #define SEMANAGE_CONF_FILE "semanage.conf"
 /* relative path names to enum semanage_paths to special files and
  * directories for the module store */
+
+#define TRUE 1
+
 enum semanage_file_defs {
 	SEMANAGE_ROOT,
 	SEMANAGE_TRANS_LOCK,
@@ -110,6 +113,8 @@ static const char *semanage_sandbox_paths[SEMANAGE_STORE_NUM_PATHS] = {
 	"/seusers.final",
 	"/users_extra",
 	"/netfilter_contexts",
+	"/file_contexts.homedirs",
+	"/disable_dontaudit",
 };
 
 /* A node used in a linked list of file contexts; used for sorting.
@@ -432,10 +437,12 @@ static int semanage_filename_select(const struct dirent *d)
  * overwrite it.  Returns 0 on success, -1 on error. */
 static int semanage_copy_file(const char *src, const char *dst, mode_t mode)
 {
-	int in, out, retval = 0, amount_read, n;
+	int in, out, retval = 0, amount_read, n, errsv = errno;
 	char tmp[PATH_MAX];
 	char buf[4192];
 
+	if (link(src,dst) == 0) return 0;
+
 	n = snprintf(tmp, PATH_MAX, "%s.tmp", dst);
 	if (n < 0 || n >= PATH_MAX)
 		return -1;
@@ -448,23 +455,32 @@ static int semanage_copy_file(const char *src, const char *dst, mode_t mode)
 		mode = S_IRUSR | S_IWUSR;
 
 	if ((out = open(tmp, O_WRONLY | O_CREAT | O_TRUNC, mode)) == -1) {
+		errsv = errno;
 		close(in);
-		return -1;
+		retval = -1;
+		goto out;
 	}
 	while (retval == 0 && (amount_read = read(in, buf, sizeof(buf))) > 0) {
-		if (write(out, buf, amount_read) != amount_read) {
+		if (write(out, buf, amount_read) < 0) {
+			errsv = errno;
 			retval = -1;
 		}
 	}
-	if (amount_read < 0)
+	if (amount_read < 0) {
+		errsv = errno;
 		retval = -1;
+	}
 	close(in);
-	if (close(out) < 0)
+	if (close(out) < 0) {
+		errsv = errno;
 		retval = -1;
+	}
 
 	if (!retval && rename(tmp, dst) == -1)
 		return -1;
 
+out:
+	errno = errsv;
 	return retval;
 }
 
@@ -538,7 +554,7 @@ int semanage_remove_directory(const char *path)
 		free(namelist[i]);
 	}
 	free(namelist);
-	if (remove(path) == -1) {
+	if (rmdir(path) == -1) {
 		return -4;
 	}
 	return 0;
@@ -553,12 +569,14 @@ int semanage_make_sandbox(semanage_handle_t * sh)
 {
 	const char *sandbox = semanage_path(SEMANAGE_TMP, SEMANAGE_TOPLEVEL);
 	struct stat buf;
+	int errsv;
 
 	if (stat(sandbox, &buf) == -1) {
 		if (errno != ENOENT) {
 			ERR(sh, "Error scanning directory %s.", sandbox);
 			return -1;
 		}
+		errno = 0;
 	} else {
 		/* remove the old sandbox */
 		if (semanage_remove_directory(sandbox) != 0) {
@@ -577,7 +595,9 @@ int semanage_make_sandbox(semanage_handle_t * sh)
 	return 0;
 
       cleanup:
+	errsv = errno;
 	semanage_remove_directory(sandbox);
+	errno = errsv;
 	return -1;
 }
 
@@ -894,14 +914,14 @@ static int semanage_exec_prog(semanage_handle_t * sh,
 
 	/* no need to use pthread_atfork() -- child will not be using
 	 * any mutexes. */
-	if ((forkval = fork()) == -1) {
+	if ((forkval = vfork()) == -1) {
 		ERR(sh, "Error while forking process.");
 		return -1;
 	} else if (forkval == 0) {
 		/* child process.  file descriptors will be closed
 		 * because they were set as close-on-exec. */
 		execve(e->path, argv, NULL);
-		exit(EXIT_FAILURE);	/* if execve() failed */
+		_exit(EXIT_FAILURE);	/* if execve() failed */
 	} else {
 		/* parent process.  wait for child to finish */
 		int status = 0;
@@ -966,16 +986,17 @@ int semanage_split_fc(semanage_handle_t * sh)
 
 	while (fgets_unlocked(buf, PATH_MAX, file_con)) {
 		if (!strncmp(buf, "HOME_DIR", 8) ||
-		    !strncmp(buf, "HOME_ROOT", 9) || strstr(buf, "ROLE")) {
+		    !strncmp(buf, "HOME_ROOT", 9) || strstr(buf, "ROLE") ||
+		    strstr(buf, "USER")) {
 			/* This contains one of the template variables, write it to homedir.template */
-			if (write(hd, buf, strlen(buf)) == 0) {
+			if (write(hd, buf, strlen(buf)) < 0) {
 				ERR(sh, "Write to %s failed.",
 				    semanage_path(SEMANAGE_TMP,
 						  SEMANAGE_HOMEDIR_TMPL));
 				goto cleanup;
 			}
 		} else {
-			if (write(fc, buf, strlen(buf)) == 0) {
+			if (write(fc, buf, strlen(buf)) < 0) {
 				ERR(sh, "Write to %s failed.",
 				    semanage_path(SEMANAGE_TMP, SEMANAGE_FC));
 				goto cleanup;
@@ -1008,14 +1029,15 @@ static int semanage_install_active(semanage_handle_t * sh)
 	const char *active_fc = semanage_path(SEMANAGE_ACTIVE, SEMANAGE_FC);
 	const char *active_fc_loc =
 	    semanage_path(SEMANAGE_ACTIVE, SEMANAGE_FC_LOCAL);
-	const char *active_hd =
-	    semanage_path(SEMANAGE_ACTIVE, SEMANAGE_HOMEDIR_TMPL);
 	const char *active_seusers =
 	    semanage_path(SEMANAGE_ACTIVE, SEMANAGE_SEUSERS);
 	const char *active_nc = semanage_path(SEMANAGE_ACTIVE, SEMANAGE_NC);
+	const char *active_fc_hd =
+	    semanage_path(SEMANAGE_ACTIVE, SEMANAGE_FC_HOMEDIRS);
 
 	const char *running_fc = selinux_file_context_path();
 	const char *running_fc_loc = selinux_file_context_local_path();
+	const char *running_fc_hd = selinux_file_context_homedir_path();
 	const char *running_hd = selinux_homedir_context_path();
 	const char *running_policy = selinux_binary_policy_path();
 	const char *running_seusers = selinux_usersconf_path();
@@ -1027,14 +1049,15 @@ static int semanage_install_active(semanage_handle_t * sh)
 	 * POLICYTYPE and should probably be done in the future. */
 	char store_fc[PATH_MAX];
 	char store_fc_loc[PATH_MAX];
-	char store_hd[PATH_MAX];
 	char store_pol[PATH_MAX];
 	char store_seusers[PATH_MAX];
 	char store_nc[PATH_MAX];
+	char store_fc_hd[PATH_MAX];
 
 	len = strlen(really_active_store);
 	running_fc += len;
 	running_fc_loc += len;
+	running_fc_hd += len;
 	running_hd += len;
 	running_policy += len;
 	running_seusers += len;
@@ -1055,10 +1078,13 @@ static int semanage_install_active(semanage_handle_t * sh)
 		goto cleanup;
 	}
 
-	snprintf(store_hd, PATH_MAX, "%s%s", storepath, running_hd);
-	if (semanage_copy_file(active_hd, store_hd, sh->conf->file_mode) == -1) {
-		ERR(sh, "Could not copy %s to %s.", active_hd, store_hd);
-		goto cleanup;
+	if (!sh->conf->disable_genhomedircon) {
+		snprintf(store_fc_hd, PATH_MAX, "%s%s", storepath, running_fc_hd);
+		if (semanage_copy_file(active_fc_hd, store_fc_hd, sh->conf->file_mode)
+			== -1) {
+			ERR(sh, "Could not copy %s to %s.", active_fc_hd, store_fc_hd);
+			goto cleanup;
+		}
 	}
 
 	snprintf(store_fc, PATH_MAX, "%s%s", storepath, running_fc);
@@ -1074,6 +1100,7 @@ static int semanage_install_active(semanage_handle_t * sh)
 		    store_fc_loc);
 		goto cleanup;
 	}
+	errno = 0;
 
 	snprintf(store_seusers, PATH_MAX, "%s%s", storepath, running_seusers);
 	if (semanage_copy_file
@@ -1083,6 +1110,7 @@ static int semanage_install_active(semanage_handle_t * sh)
 		    store_seusers);
 		goto cleanup;
 	}
+	errno = 0;
 
 	snprintf(store_nc, PATH_MAX, "%s%s", storepath, running_nc);
 	if (semanage_copy_file(active_nc, store_nc, sh->conf->file_mode) == -1
@@ -1090,6 +1118,7 @@ static int semanage_install_active(semanage_handle_t * sh)
 		ERR(sh, "Could not copy %s to %s.", active_nc, store_nc);
 		goto cleanup;
 	}
+	errno = 0;
 
 	if (!sh->do_reload)
 		goto skip_reload;
@@ -1111,8 +1140,10 @@ static int semanage_install_active(semanage_handle_t * sh)
 			goto skip_reload;
 		}
 	} else if (errno == ENOENT &&
-		   strcmp(really_active_store, storepath) != 0)
+		   strcmp(really_active_store, storepath) != 0) {
+		errno = 0;
 		goto skip_reload;
+	}
 
 	if (semanage_reload_policy(sh)) {
 		goto cleanup;
@@ -1197,6 +1228,10 @@ static int semanage_commit_sandbox(semanage_handle_t * sh)
 		retval = -1;
 		goto cleanup;
 	}
+
+	/* clean up some files from the sandbox before install */
+	/* remove homedir_template from sandbox */
+
 	if (rename(sandbox, active) == -1) {
 		ERR(sh, "Error while renaming %s to %s.", sandbox, active);
 		/* note that if an error occurs during the next
@@ -1212,6 +1247,7 @@ static int semanage_commit_sandbox(semanage_handle_t * sh)
 		/* note that if an error occurs during the next three
 		 * function then the store will be left in an
 		 * inconsistent state */
+		int errsv = errno;
 		if (rename(active, sandbox) < 0)
 			ERR(sh, "Error while renaming %s back to %s.", active,
 			    sandbox);
@@ -1220,16 +1256,19 @@ static int semanage_commit_sandbox(semanage_handle_t * sh)
 			    active);
 		else
 			semanage_install_active(sh);
+		errno = errsv;
 		retval = -1;
 		goto cleanup;
 	}
 
 	if (!sh->conf->save_previous) {
+		int errsv = errno;
 		retval = semanage_remove_directory(backup);
 		if (retval < 0) {
 			ERR(sh, "Could not delete previous directory %s.", backup);
 			goto cleanup;
 		}
+		errno = errsv;
 	}
 
       cleanup:
@@ -1262,13 +1301,6 @@ int semanage_install_sandbox(semanage_handle_t * sh)
 		goto cleanup;
 	}
 
-	if ((retval =
-	     semanage_exec_prog(sh, sh->conf->genhomedircon,
-				sh->conf->store_path, "")) != 0) {
-		ERR(sh, "genhomedircon returned error code %d.", retval);
-		goto cleanup;
-	}
-
 	retval = commit_num;
 
       cleanup:
@@ -1391,22 +1423,26 @@ int semanage_get_active_lock(semanage_handle_t * sh)
  * there. */
 void semanage_release_trans_lock(semanage_handle_t * sh)
 {
+	int errsv = errno;
 	if (sh->u.direct.translock_file_fd >= 0) {
 		flock(sh->u.direct.translock_file_fd, LOCK_UN);
 		close(sh->u.direct.translock_file_fd);
 		sh->u.direct.translock_file_fd = -1;
 	}
+	errno = errsv;
 }
 
 /* Releases the read lock.  Does nothing if there was not one already
  * there. */
 void semanage_release_active_lock(semanage_handle_t * sh)
 {
+	int errsv = errno;
 	if (sh->u.direct.activelock_file_fd >= 0) {
 		flock(sh->u.direct.activelock_file_fd, LOCK_UN);
 		close(sh->u.direct.activelock_file_fd);
 		sh->u.direct.activelock_file_fd = -1;
 	}
+	errno = errsv;
 }
 
 /* Read the current commit number from the commit number file which
@@ -1432,6 +1468,7 @@ int semanage_direct_get_serial(semanage_handle_t * sh)
 		if (errno == ENOENT) {
 			/* the commit number file does not exist yet,
 			 * so assume that the number is 0 */
+			errno = 0;
 			return 0;
 		} else {
 			ERR(sh, "Could not open commit number file %s.",
@@ -1489,16 +1526,30 @@ static int semanage_load_module(semanage_handle_t * sh, const char *filename,
 		ERR(sh, "Could not open module file %s for reading.", filename);
 		goto cleanup;
 	}
+	ssize_t size;
+	char *data = NULL;
+
+	if ((size = bunzip(fp, &data)) > 0) {
+		fclose(fp);
+		fp = fmemopen(data, size, "rb");
+		if (!fp) {
+			ERR(sh, "Out of memory!");
+			goto cleanup;
+		}
+	}
+	rewind(fp);
 	__fsetlocking(fp, FSETLOCKING_BYCALLER);
 	sepol_policy_file_set_fp(pf, fp);
 	sepol_policy_file_set_handle(pf, sh->sepolh);
 	if (sepol_module_package_read(*package, pf, 0) == -1) {
 		ERR(sh, "Error while reading from module file %s.", filename);
 		fclose(fp);
+		free(data);
 		goto cleanup;
 	}
 	sepol_policy_file_free(pf);
 	fclose(fp);
+	free(data);
 	return retval;
 
       cleanup:
@@ -1575,6 +1626,41 @@ int semanage_link_sandbox(semanage_handle_t * sh,
 	return retval;
 }
 
+/* Links only the base module within the sandbox into the base module.
+ * '*base' will point to the module package that contains everything
+ * linked together (caller must call sepol_module_package_destroy() on
+ * it afterwards).  '*base' will be set to NULL upon entering this
+ * function.  Returns 0 on success, -1 on error.
+ */
+int semanage_link_base(semanage_handle_t * sh,
+			  sepol_module_package_t ** base)
+{
+	const char *base_filename = NULL;
+	int retval = -1;
+
+	*base = NULL;
+
+	/* first make sure that base module is readable */
+	if ((base_filename =
+	     semanage_path(SEMANAGE_TMP, SEMANAGE_BASE)) == NULL) {
+		goto cleanup;
+	}
+	if (access(base_filename, R_OK) == -1) {
+		ERR(sh, "Could not access sandbox base file %s.",
+		    base_filename);
+		goto cleanup;
+	}
+
+	if (semanage_load_module(sh, base_filename, base) == -1) {
+		goto cleanup;
+	}
+
+	retval = 0;
+
+      cleanup:
+	return retval;
+}
+
 /* 
  * Expands the policy contained within *base 
  */
@@ -1590,6 +1676,8 @@ int semanage_expand_sandbox(semanage_handle_t * sh,
 	if (sepol_policydb_create(&out))
 		goto err;
 
+	sepol_set_expand_consume_base(sh->sepolh, 1);
+
 	if (sepol_expand_module(sh->sepolh,
 				sepol_module_package_get_policy(base), out, 0,
 				expand_check)
@@ -1601,6 +1689,8 @@ int semanage_expand_sandbox(semanage_handle_t * sh,
 		ERR(sh, "Unknown/Invalid policy version %d.", policyvers);
 		goto err;
 	}
+	if (sh->conf->handle_unknown >= 0)
+		sepol_policydb_set_handle_unknown(out, sh->conf->handle_unknown);
 
 	*policydb = out;
 	return STATUS_SUCCESS;
@@ -1610,6 +1700,47 @@ int semanage_expand_sandbox(semanage_handle_t * sh,
 	return STATUS_ERR;
 }
 
+/**
+ * Read the policy from the sandbox (kernel)
+ */
+int semanage_read_policydb(semanage_handle_t * sh, sepol_policydb_t * in)
+{
+
+	int retval = STATUS_ERR;
+	const char *kernel_filename = NULL;
+	struct sepol_policy_file *pf = NULL;
+	FILE *infile = NULL;
+
+	if ((kernel_filename =
+	     semanage_path(SEMANAGE_ACTIVE, SEMANAGE_KERNEL)) == NULL) {
+		goto cleanup;
+	}
+	if ((infile = fopen(kernel_filename, "r")) == NULL) {
+		ERR(sh, "Could not open kernel policy %s for reading.",
+		    kernel_filename);
+		goto cleanup;
+	}
+	__fsetlocking(infile, FSETLOCKING_BYCALLER);
+	if (sepol_policy_file_create(&pf)) {
+		ERR(sh, "Out of memory!");
+		goto cleanup;
+	}
+	sepol_policy_file_set_fp(pf, infile);
+	sepol_policy_file_set_handle(pf, sh->sepolh);
+	if (sepol_policydb_read(in, pf) == -1) {
+		ERR(sh, "Error while reading kernel policy from %s.",
+		    kernel_filename);
+		goto cleanup;
+	}
+	retval = STATUS_SUCCESS;
+
+      cleanup:
+	if (infile != NULL) {
+		fclose(infile);
+	}
+	sepol_policy_file_free(pf);
+	return retval;
+}
 /**
  * Writes the final policy to the sandbox (kernel)
  */
